"""[User Facing] Root package for tournamentcontrol project.

This package contains Django applications specific to the tournamentcontrol
installation and is not intended for reuse outside this project.
"""
